import utm
from obspy import read_events, read_inventory
from data_gen_01d import data_gen
from vmodel_gen_01d import vmodel_gen
from trmloc_scenario import trmloc_scenario

BOUNDARY_EXTENSION = 0.005  # ~500m extension for min/max coordinates

LOG_FILE_NAME = "log.txt"
OUT_FILE_NAME = "out.txt"
RES_FILE_NAME = "res.txt"
XCP_FILE_NAME = "xcp.txt"
INPUT_PAR_FILE_NAME = "inputPar.txt"
MAX_LONGITUDE_IN_ZONE = 6 - 1e-9
MIN_LONGITUDE_IN_ZONE = 0

def main(network_inventory_path, picks_path, velocity_model_file_path, latitude, longitude, depth, event_name, fvi,
         fdi, fpo, err, cme, cmz, ia, ip, cp, omp, fi, fh, cp_min, cp_max, cpi, cpf, ftm, fit, fsc, fec, m, e, xstep,
         ystep, z1, z2, zstep, log_file_name=LOG_FILE_NAME, out_file_name=OUT_FILE_NAME, res_file_name=RES_FILE_NAME,
         xcp_file_name=XCP_FILE_NAME, input_par_file_name=INPUT_PAR_FILE_NAME
         ):
    inventory = read_inventory(network_inventory_path)
    picks = read_events(picks_path).events[0].picks
    waveforms_with_p_points = get_waveforms_with_points([pick for pick in picks if pick.phase_hint[0] == "P"])
    pickedSeismogramsNo = len(waveforms_with_p_points.keys())
    if pickedSeismogramsNo < 3:
        raise ValueError(
            f"Picks have to be present on at least 3 seismograms. Found picks only for {pickedSeismogramsNo}.")

    data_file_name = prepare_data_file(
        waveforms_with_p_points, inventory, latitude, longitude
    )
    vmodel_file_name = prepare_vmodel_file(
        inventory, velocity_model_file_path, latitude, longitude, xstep, ystep, z1, z2, zstep
    )

    params_file_name = prepare_params_file( data_file_name, vmodel_file_name, latitude, longitude, depth, event_name,
                                            fvi, fdi, fpo, err, cme, cmz, ia, ip, cp, omp, fi, fh, cp_min, cp_max, cpi, cpf, ftm, fit, fsc, fec, m, e,
                                            log_file_name, out_file_name, res_file_name, xcp_file_name, input_par_file_name
                                            )

    return params_file_name


def prepare_data_file(seismograms_with_points, inventory, latitude, longitude):
    utm_zone = utm.from_latlon(latitude, longitude)[2:4]
    station_data = {"station_x": [], "station_y": [], "station_elevations": [], "pick_times": []}

    for points in seismograms_with_points.values():
        channel = find_channel(points[0].waveform_id, inventory)
        if channel is None:
            raise ValueError(f"Channel {points[0].waveform_id} not found")
        station_utm_coords = convert_to_utm(channel.latitude, channel.longitude, utm_zone[0], utm_zone[1])

        if len(points) != 1:
            raise ValueError(
                f"Seismogram {points[0].waveform_id} contains more than one picked point."
                " This would cause the application to produce inconsistent results"
            )

        for point in points:
            station_data["station_x"].append(station_utm_coords[0])
            station_data["station_y"].append(station_utm_coords[1])
            station_data["station_elevations"].append(-channel.elevation)
            station_data["pick_times"].append(point.time)

    data_file_path = "data.txt"
    data_gen(
        data_file_path,
        station_data["station_x"],
        station_data["station_y"],
        station_data["station_elevations"],
        station_data["pick_times"],
    )
    return data_file_path


def prepare_vmodel_file(inventory, velocity_model_file_path, latitude, longitude, xstep, ystep, z1, z2, zstep):
    utm_zone = utm.from_latlon(latitude, longitude)[2:4]
    min_coords, max_coords = get_min_max_coordinates(inventory)
    min_utm = convert_to_utm(min_coords["latitude"], min_coords["longitude"], utm_zone[0], utm_zone[1])
    max_utm = convert_to_utm(max_coords["latitude"], max_coords["longitude"], utm_zone[0], utm_zone[1])

    vmodel_file_path = "VeloMod.txt"
    vmodel_gen(
        vmodel_file_path, "Velocity model", 0,  # Struc	0 or 1 [default set to 0. 1 is not supported yet]
        round(min_utm[0]), round(max_utm[0]), xstep,
        round(min_utm[1]), round(max_utm[1]), ystep,
        z1, z2, zstep, velocity_model_file_path
    )
    return vmodel_file_path


def prepare_params_file( data_file_name, vmodel_file_name, latitude, longitude, depth, event_name,
                         fvi, fdi, fpo, err, cme, cmz, ia, ip, cp, omp, fi, fh, cp_min, cp_max, cpi, cpf, ftm, fit, fsc, fec, m, e,
                         log_file_name, out_file_name, res_file_name, xcp_file_name, input_par_file_name
                         ):
    utm_zone = utm.from_latlon(latitude, longitude)[2:4]
    event_utm = convert_to_utm(latitude, longitude, utm_zone[0], utm_zone[1])
    trmloc_scenario(
        input_par_file_name,
        data_file_name,
        vmodel_file_name,
        "",  # cp list file inactive
        log_file_name, out_file_name, res_file_name, xcp_file_name,
        event_name, fvi, fdi, fpo, err, cme, cmz, ia, event_utm[0], event_utm[1], depth * 1000, ip,
        cp, omp, fi, fh, cp_min, cp_max, cpi, cpf, ftm, fit, fsc, fec, m, e
    )
    return input_par_file_name


def get_min_max_coordinates(inventory):
    channels = get_all_channels(inventory)
    latitudes = [channel.latitude for channel in channels]
    longitudes = [channel.longitude for channel in channels]

    # Extend boundaries by 0.005 degrees (~500 meters) in any direction
    min_coords = {"latitude": min(latitudes) - BOUNDARY_EXTENSION, "longitude": min(longitudes) - BOUNDARY_EXTENSION}
    max_coords = {"latitude": max(latitudes) + BOUNDARY_EXTENSION, "longitude": max(longitudes) + BOUNDARY_EXTENSION}
    return min_coords, max_coords

def get_waveforms_with_points(picks):
    stack = []
    for pick in picks:
        stack.append(pick)

    grouped_picks = {}
    while stack:
        pick = stack.pop()
        if pick.phase_hint.endswith("e"):
            stack.pop()
        else:
            waveform_id_string = stringify_waveform_id(pick.waveform_id)
            if waveform_id_string not in grouped_picks:
                grouped_picks[waveform_id_string] = []
            grouped_picks[waveform_id_string].append(pick)

    return grouped_picks

def find_channel(waveform_id, networks):
    return next(
        (
            channel
            for network in networks if network.code == waveform_id.network_code
            for station in network.stations if station.code == waveform_id.station_code
            for channel in station.channels
            if channel.location_code == waveform_id.location_code
                                       and channel.code == waveform_id.channel_code
        ),
        None
    )

def get_all_channels(networks):
    return [channel for network in networks for station in network.stations for channel in station.channels]

def convert_to_utm(latitude, longitude, target_longitude_zone, target_latitude_zone):

    easting, northing, zone_number, zone_letter = utm.from_latlon(latitude, longitude)

    if (zone_letter >= 'N' and target_latitude_zone < 'N') or (zone_letter < 'N' and target_latitude_zone >= 'N'):
        raise ValueError(
            f"Points must be on the same hemisphere (N/S). Offending point latitude:{latitude}, longitude: {longitude}"
        )

    POLAR_ZONE_CODES = {'A', 'B', 'Y', 'Z'}
    if zone_letter in POLAR_ZONE_CODES:
        raise ValueError(
            f"Polar regions not supported. Offending point latitude:{latitude}, longitude: {longitude}"
        )

    if zone_number != target_longitude_zone:
        if abs(target_longitude_zone - zone_number) > 2:
            raise ValueError(
                f"Point latitude:{latitude}, longitude: {longitude} is too far from requested UTM zone {target_longitude_zone}"
            )
        # Recalculate to the requested UTM zone if necessary
        easting, northing, _, _ = recalculate_to_zone(
            [easting, northing, zone_number, zone_letter], target_longitude_zone, latitude
        )

    return [easting, northing]

def recalculate_to_zone(utm_coords, target_longitude_zone, latitude):
    zone_difference = get_longitude_zone_difference(utm_coords[2], target_longitude_zone)

    min_easting, _, _, _ = utm.from_latlon(latitude, MIN_LONGITUDE_IN_ZONE)
    max_easting, _, _, _ = utm.from_latlon(latitude, MAX_LONGITUDE_IN_ZONE)

    shift = zone_difference * (max_easting - min_easting)
    adjusted_easting = utm_coords[0] + shift

    return [adjusted_easting, utm_coords[1], utm_coords[2], utm_coords[3]]

def get_longitude_zone_difference(longitude_zone1, longitude_zone2):
    difference = longitude_zone1 - longitude_zone2
    if difference < -30:
        return -60 - difference
    if difference <= 30:
        return difference
    return 60 - difference

def stringify_waveform_id(waveform_id):
    return f"{waveform_id.network_code}{waveform_id.station_code}{waveform_id.location_code}{waveform_id.channel_code}"