import numpy as np
from scipy.io import loadmat

def vmodel_gen(P, Discr, Struc, x1, x2, xstep, y1, y2, ystep, z1, z2, zstep, velocity_model_file_path):
    """
    Generate a 1D velocity model and save it to a file.
    
    Parameters:
    - P: filename for output.
    - Discr: description of the model, typed by the user [string up to 255 characters].
    - Struc: structure flag (0 or 1).
    - x1, x2, xstep: minimum, maximum values and step for generation of the X-Axis.
    - y1, y2, ystep: minimum, maximum values and step for generation of the Y-Axis.
    - z1, z2, zstep: minimum, maximum values and step for generation of the Z-Axis.
    - velocity_model_file_path: path to matlab file with P-wave velocity values.
        In the current version 2 options are supported:
        1) 0d model: the user types a single value
        2) 1d model: a Nx2 array is inserted, corresponding to the depth and
             velocity of N layers. e.g.
                         200    5800       (depth - velocity)
                         800    6200
                       1500     6500
    """
    data = loadmat(velocity_model_file_path)

    data = np.column_stack((
        np.array(data['d']['Depth'][0, 0]).flatten(),
        np.array(data['d']['Vp'][0, 0]).flatten()
    ))
    data = data * 1000

    szx = int(np.ceil((x2 - x1) / xstep))
    szy = int(np.ceil((y2 - y1) / ystep))
    szz = int(np.ceil((z2 - z1) / zstep))
    z = np.arange(z1, z2 + 1, zstep)

    if data.shape[0] == 1:
        vz = np.full(z.shape, data[0, 1])
        Dimens = 0
    elif data.shape[0] > 1:
        data = np.vstack((data, [1000000, data[-1, 1]]))
        vz = np.zeros_like(z)                   
        for j in range(1, len(data)):
            f1 = (z < data[j, 0]) & (z >= data[j - 1, 0])
            vz[f1] = data[j - 1, 1]
        vz[z >= data[-1, 0]] = data[-1, 1]
        Dimens = 1
    else:
        raise ValueError("Complex velocity models are not supported.")

    if np.any(vz == 0):
        vz[vz == 0] = vz[vz != 0][0]

    with open(P, 'w') as file:
        file.write("# TNF 1.0 \n")
        file.write(f"# DIMENSION {Dimens} \n")
        file.write("# FORMAT ASCII \n")
        file.write(f"# DESCRIPTION {Discr} \n")
        file.write(f"# STRUCTURE {Struc} \n")
        file.write(f"# SIZE {szx} {szy} {szz} {szx * szy * szz} \n")
        file.write(f"# XAXIS {x1} {x2} {xstep} \n")
        file.write(f"# YAXIS {y1} {y2} {ystep} \n")
        file.write(f"# ZAXIS {z1} {z2} {zstep} \n")
        np.savetxt(file, vz, fmt="%d")